/// Bazel build system implementation.
pub mod bazel;
/// GNOME build system implementation.
pub mod gnome;
/// Go build system implementation.
pub mod go;
/// Haskell build system implementation.
pub mod haskell;
/// Java build system implementation.
pub mod java;
/// Make build system implementation.
pub mod make;
/// Meson build system implementation.
pub mod meson;
/// Node.js build system implementation.
pub mod node;
/// Octave build system implementation.
pub mod octave;
/// Perl build system implementation.
pub mod perl;
/// Python build system implementation.
pub mod python;
/// R build system implementation.
pub mod r;
/// Ruby build system implementation.
pub mod ruby;
/// Rust build system implementation.
pub mod rust;
/// Waf build system implementation.
pub mod waf;
